package clinical.web.services;

import java.math.BigDecimal;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import clinical.server.dao.ConfParamsDAO;
import clinical.server.dao.ConfParamsGroupDAO;
import clinical.server.vo.ConfParams;
import clinical.server.vo.ConfParamsGroup;
import clinical.server.vo.Databaseuser;
import clinical.utils.Assertion;
import clinical.utils.GenUtils;
import clinical.web.Constants;
import clinical.web.DAOFactory;
import clinical.web.IAppConfigService;
import clinical.web.ISQLDialect;
import clinical.web.ISequenceHelper;
import clinical.web.MinimalServiceFactory;
import clinical.web.common.IDBPoolService;
import clinical.web.common.query.TSQLProcessor;
import clinical.web.exception.BaseException;

/**
 * @author I. Burak Ozyurt
 * @version $Id: AppConfigService.java 384 2011-06-23 00:15:36Z bozyurt $
 */
public class AppConfigService implements IAppConfigService {
	private Map<String, ConfParamsGroup> paramGroupMap = Collections
			.synchronizedMap(new HashMap<String, ConfParamsGroup>(11));
	private Map<BigDecimal, ConfParamsGroup> pgByIdMap = Collections
			.synchronizedMap(new HashMap<BigDecimal, ConfParamsGroup>(11));
	private Map<String, ConfParams> paramMap = Collections
			.synchronizedMap(new HashMap<String, ConfParams>());

	// private ConcurrentMap<BigDecimal, ConfParamsGroup> pgByIdMap = new
	// ConcurrentHashMap<BigDecimal, ConfParamsGroup>(
	// 11);
	// private ConcurrentMap<String, ConfParams> paramMap = new
	// ConcurrentHashMap<String, ConfParams>();

	private String dbID;
	private IDBPoolService poolService;
	private static AppConfigService instance = null;
	private Log log = LogFactory.getLog(AppConfigService.class);

	private AppConfigService(IDBPoolService poolService, String dbID)
			throws BaseException {
		this.poolService = poolService;
		this.dbID = dbID;
		Connection con = null;
		try {
			con = poolService.getConnection(Constants.ADMIN_USER);
			loadAllParams(con);
		} catch (Exception e) {
			if (e instanceof BaseException) {
				throw (BaseException) e;
			}
			throw new BaseException(e);
		} finally {
			poolService.releaseConnection(Constants.ADMIN_USER, con);
		}
	}

	public static synchronized AppConfigService getInstance(
			IDBPoolService poolService, String dbID) throws BaseException {
		if (instance == null) {
			instance = new AppConfigService(poolService, dbID);
		}
		return instance;
	}

	public void reload() throws BaseException {
		Connection con = null;
		try {
			con = poolService.getConnection(Constants.ADMIN_USER);
			synchronized (this) {
				this.paramMap.clear();
				this.paramGroupMap.clear();
				this.pgByIdMap.clear();
				loadAllParams(con);
			}
		} catch (Exception e) {
			if (e instanceof BaseException) {
				throw (BaseException) e;
			}
			throw new BaseException(e);
		} finally {
			poolService.releaseConnection(Constants.ADMIN_USER, con);
		}
	}

	/**
	 * 
	 * @return the singleton AppConfigService
	 * @throws BaseException
	 *             if not created before
	 */
	public static synchronized AppConfigService getInstance()
			throws BaseException {
		if (instance == null)
			throw new BaseException(
					"Application Configuration Service is not initialized!");
		return instance;
	}

	public ConfParamsGroup addParameterGroup(String groupName,
			String description) throws Exception {
		Connection con = null;
		try {
			con = poolService.getConnection(Constants.ADMIN_USER);
			con.setAutoCommit(false);
			ConfParamsGroup pg = addParameterGroup(con, groupName, description);
			con.commit();
			paramGroupMap.put(pg.getGroupName(), pg);
			pgByIdMap.put(pg.getUniqueId(), pg);
			return pg;
		} catch (Exception x) {
			handleErrorAndRollBack(con, "", x);
			return null;
		} finally {
			poolService.releaseConnection(Constants.ADMIN_USER, con);
		}
	}

	public void removeParameterGroup(String groupName) throws Exception {
		Connection con = null;
		try {
			con = poolService.getConnection(Constants.ADMIN_USER);
			con.setAutoCommit(false);
			removeParameterGroup(con, groupName);
			con.commit();
			synchronized (this) {
				paramGroupMap.remove(groupName);
				ConfParamsGroup thePG = null;
				for (ConfParamsGroup pg : pgByIdMap.values()) {
					if (pg.getGroupName().equals(groupName)) {
						thePG = pg;
						break;
					}
				}
				pgByIdMap.remove(thePG.getUniqueId());
			}
		} catch (Exception x) {
			handleErrorAndRollBack(con, "", x);
		} finally {
			poolService.releaseConnection(Constants.ADMIN_USER, con);
		}
	}

	public void addParameter(String name, String value, String type,
			String description, String groupName) throws Exception {
		Connection con = null;
		try {
			con = poolService.getConnection(Constants.ADMIN_USER);
			con.setAutoCommit(false);
			ConfParams p = addParameter(con, name, value, type, description,
					groupName);
			con.commit();
			paramMap.put(p.getName(), p);
		} catch (Exception x) {
			handleErrorAndRollBack(con, "", x);
		} finally {
			poolService.releaseConnection(Constants.ADMIN_USER, con);
		}
	}

	public void updateParameter(String name, String value) throws Exception {
		Connection con = null;
		try {
			con = poolService.getConnection(Constants.ADMIN_USER);
			con.setAutoCommit(false);
			updateParameter(con, name, value);
			con.commit();
			synchronized (this) {
				ConfParams p = paramMap.get(name);
				p.setValue(value);
			}
		} catch (Exception x) {
			handleErrorAndRollBack(con, "", x);
		} finally {
			poolService.releaseConnection(Constants.ADMIN_USER, con);
		}
	}

	public void removeParameter(String name) throws Exception {
		Connection con = null;
		try {
			con = poolService.getConnection(Constants.ADMIN_USER);
			con.setAutoCommit(false);
			removeParameter(con, name);
			con.commit();
			paramMap.remove(name);
		} catch (Exception x) {
			handleErrorAndRollBack(con, "", x);
		} finally {
			poolService.releaseConnection(Constants.ADMIN_USER, con);
		}
	}

	public ConfParams getParam(String name) {
		return paramMap.get(name);
	}

	public String getParamValue(String name) {
		ConfParams cp = paramMap.get(name);
		if (cp == null || cp.getValue() == null)
			return null;
		return cp.getValue().trim();
	}

	public String[] getParamValues(String name) {
		List<ParamOrder> poList = new ArrayList<ParamOrder>(10);
		synchronized (this) {
			for (String pn : paramMap.keySet()) {
				if (pn.startsWith(name)) {
					ConfParams cp = paramMap.get(pn);
					String s = pn.substring(name.length() + 1);
					int idx = GenUtils.toInt(s, -1);
					if (idx != -1) {
						ParamOrder po = new ParamOrder(idx, cp.getValue());
						poList.add(po);
					}
				}
			}
		}
		Collections.sort(poList, new Comparator<ParamOrder>() {
			public int compare(ParamOrder o1, ParamOrder o2) {
				return o1.idx - o2.idx;
			}
		});
		String[] values = new String[poList.size()];
		int i = 0;
		for (ParamOrder po : poList) {
			values[i++] = po.value;
		}
		return values;
	}

	class ParamOrder {
		int idx;
		String value;

		public ParamOrder(int idx, String value) {
			this.idx = idx;
			this.value = value;
		}
	}// ;

	public synchronized ConfParamsGroup getGroupForParam(String name) {
		ConfParams p = paramMap.get(name);
		if (p == null)
			return null;
		return pgByIdMap.get(p.getParamGroupId());
	}

	public synchronized List<ConfParamsGroup> getAllParamGroups() {
		return new ArrayList<ConfParamsGroup>(pgByIdMap.values());
	}

	public synchronized List<ConfParams> getAllParams() {
		return new ArrayList<ConfParams>(paramMap.values());
	}

	public ConfParamsGroup getParamGroup(String groupName) {
		return paramGroupMap.get(groupName);
	}

	protected ConfParamsGroup addParameterGroup(Connection con,
			String groupName, String description) throws Exception {
		ISequenceHelper sequenceHelper = MinimalServiceFactory
				.getSequenceHelper(dbID);
		ConfParamsGroupDAO pgDAO = DAOFactory.createConfParamsGroupDAO(dbID);
		Databaseuser d = getDatabaseUser(con);
		ConfParamsGroup pg = new ConfParamsGroup();
		pg.setGroupName(groupName);
		pg.setDescription(description);
		pg.setOwner(d.getUniqueid());
		pg.setModUser(d.getUniqueid());
		pg.setModTime(new Date());
		BigDecimal uniqueId = sequenceHelper.getNextUID(con,
				"nc_conf_params_group", "uniqueId");
		pg.setUniqueId(uniqueId);
		pgDAO.insert(con, pg);
		return pg;
	}

	protected void removeParameterGroup(Connection con, String groupName)
			throws Exception {
		ConfParamsGroupDAO pgDAO = DAOFactory.createConfParamsGroupDAO(dbID);
		ConfParamsDAO pDAO = DAOFactory.createConfParamsDAO(dbID);

		ConfParamsGroup cr = new ConfParamsGroup();
		cr.setGroupName(groupName);
		List<ConfParamsGroup> pgList = pgDAO.find(con, cr);
		if (pgList.isEmpty()) {
			return;
		}
		ConfParamsGroup pg = pgList.get(0);
		ConfParams pcr = new ConfParams();
		pcr.setParamGroupId(pg.getUniqueId());
		pDAO.delete(con, pcr);
		pgDAO.delete(con, cr);
	}

	protected ConfParams addParameter(Connection con, String name,
			String value, String type, String description, String groupName)
			throws Exception {
		ConfParamsGroupDAO pgDAO = DAOFactory.createConfParamsGroupDAO(dbID);
		ConfParamsDAO pDAO = DAOFactory.createConfParamsDAO(dbID);

		ConfParamsGroup pgCR = new ConfParamsGroup();
		pgCR.setGroupName(groupName);
		List<ConfParamsGroup> pgList = pgDAO.find(con, pgCR);
		if (pgList.isEmpty() || pgList.size() != 1) {
			throw new Exception(
					"Nonexistent or not unique parameter group name:"
							+ groupName);
		}
		ConfParamsGroup pg = pgList.get(0);
		ConfParams p = new ConfParams();
		p.setName(name);
		p.setValue(value);
		p.setType(type);
		p.setDescription(description);
		p.setParamGroupId(pg.getUniqueId());
		Databaseuser d = getDatabaseUser(con);
		p.setOwner(d.getUniqueid());
		p.setModUser(d.getUniqueid());
		p.setModTime(new Date());
		pDAO.insert(con, p);
		return p;
	}

	public void updateParameter(Connection con, String name, String value)
			throws Exception {
		ConfParamsDAO pDAO = DAOFactory.createConfParamsDAO(dbID);
		ConfParams cr = new ConfParams();
		cr.setName(name);
		List<ConfParams> pList = pDAO.find(con, cr);
		if (!pList.isEmpty()) {
			ConfParams p = new ConfParams();
			p.setValue(value);
			pDAO.update(con, p, cr);
		}
	}

	protected void removeParameter(Connection con, String name)
			throws Exception {
		ConfParamsDAO pDAO = DAOFactory.createConfParamsDAO(dbID);
		ConfParams cr = new ConfParams();
		cr.setName(name);
		pDAO.delete(con, cr);
	}

	protected void loadAllParams(Connection con) throws Exception {
		ConfParamsDAO pDAO = DAOFactory.createConfParamsDAO(dbID);
		ConfParamsGroupDAO pgDAO = DAOFactory.createConfParamsGroupDAO(dbID);

		List<ConfParamsGroup> pgList = pgDAO.find(con, new ConfParamsGroup());
		List<ConfParams> pList = pDAO.find(con, new ConfParams());

		synchronized (this) {
			for (ConfParams p : pList) {
				paramMap.put(p.getName(), p);
			}

			for (ConfParamsGroup pg : pgList) {
				paramGroupMap.put(pg.getGroupName(), pg);
				pgByIdMap.put(pg.getUniqueId(), pg);
			}
		}
	}

	protected Databaseuser getDatabaseUser(Connection con) throws Exception {
		ISQLDialect sqlDialect = MinimalServiceFactory.getSQLDialect(dbID);
		TSQLProcessor tsp = new TSQLProcessor(sqlDialect);
		List<?> results = tsp
				.executeQuery(con,
						"select u.* from Databaseuser as u where u.name ='ADMIN' and u.isgroup = false");
		Assertion.assertFalse(results.isEmpty());
		Databaseuser dbUser = (Databaseuser) results.get(0);
		return dbUser;
	}

	protected void handleErrorAndRollBack(Connection con, String msg,
			Exception x) throws BaseException {
		if (con != null) {
			try {
				con.rollback();
			} catch (SQLException se) {
				log.error("", se);
				throw new BaseException(se);
			}
		}
		log.error(msg, x);
		if (x instanceof BaseException)
			throw (BaseException) x;
		else
			throw new BaseException(x);
	}
}
